from __future__ import annotations

from typing import Iterable, Union

from aioinflux import InfluxDBClient  # type: ignore
from loguru import logger
from pandas import DataFrame  # type: ignore

from illuminate.exceptions import BasicExporterException
from illuminate.exporter import Exporter


class InfluxDBExporter(Exporter):
    """
    InfluxDBExporter class, writes data to InfluxDB database asynchronously.
    Inherits Exporter class and implements export method.

    Each InfluxDBExporter object is responsible for a single transaction with a
    single database. Attribute name is used to acquire database session object
    from Manager's sessions attribute.

    Constructor kwargs will be passed to client write method. For more
    information on write method, visit:
    https://aioinflux.readthedocs.io/en/stable/api.html

    Supported write data objects:
        - DataFrame with DatetimeIndex
        - Dict containing the keys: measurement, time, tags,
        fields
        - String (str or bytes) properly formatted in InfluxDB’s line protocol
        - An iterable of one of the above

    """

    name: str
    """
    InfluxDB database name selector used to get database session object from
    Manager.sessions attribute.
    """

    def __init__(
        self,
        points: Union[
            Union[DataFrame, dict, str, bytes],
            Iterable[Union[DataFrame, dict, str, bytes]],
        ],
        *args,
        **kwargs,
    ):
        """
        InfluxDBExporter's __init__ method.

        :param points: Supported data objects
        """
        self.params = kwargs
        self.points = points

    async def export(self, session: InfluxDBClient, *args, **kwargs) -> None:
        """
        Writes data to Influxdb asynchronously.

        :param session: InfluxDBClient object
        :return: None
        :raises BasicExporterException:
        """
        try:
            await session.write(self.points, **self.params)
        except Exception as exception:
            logger.warning(
                f'{self}.export(session="{session}") -> {exception}'
            )
            raise BasicExporterException
        logger.success(f'{self}.export(session="{session}")')

    def __repr__(self):
        """
        InfluxDBExporter's __repr__ method.

        :return: String representation of an instance
        """
        return f"InfluxDBExporter(points={self.points})"



from __future__ import annotations

import functools
from typing import Callable

from loguru import logger
from sqlalchemy.exc import NoSuchModuleError

from illuminate.exceptions import BasicManagerException
from illuminate.manager import Assistant


def adapt(command: str) -> Callable:
    """
    Adapts Manager's static methods to accept cli arguments.

    :param command: Command string
    :return: Manager's static method wrapper
    """

    def decorator(func: Callable) -> Callable:
        """
        Outer wrapper responsible for control flow.

        :param func: The function to be adapted
        :return: The wrapped function with adapted arguments
        """

        if command == "populate":

            @functools.wraps(func)
            def wrapper(
                fixtures: tuple[str], selector: str, url: str, *args, **kwargs
            ) -> None:
                """
                Adapts Manager's db_populate method to accept cli arguments.

                :param fixtures: Tuple of fixture files
                :param selector: Database name in settings.py module
                :param url: SQLAlchemy Database URL
                :return: None
                """

                try:
                    func(
                        fixtures,
                        selector,
                        url or Assistant._provide_db_url(selector),
                        *args,
                        **kwargs,
                    )
                except NoSuchModuleError:
                    raise BasicManagerException(
                        "Command populate can only be performed on SQL "
                        f"database, {selector} is not supported SQL database"
                    )

        elif command == "revision":

            @functools.wraps(func)
            def wrapper(
                path: str,
                revision: str,
                selector: str,
                url: str,
                *args,
                **kwargs,
            ) -> None:
                """
                Adapts Manager's db_revision method to accept cli arguments.

                :param path: Migration directory path
                :param revision: Parent revision
                :param selector: Database name in settings.py module
                :param url: SQLAlchemy Database URL
                :return: None
                """
                try:
                    config = Assistant.provide_alembic_config(
                        path, selector, url
                    )
                    func(config, revision, *args, **kwargs)
                except NoSuchModuleError:
                    raise BasicManagerException(
                        "Command revision can only be performed on SQL "
                        f"database, {selector} is not supported SQL database"
                    )

        elif command == "upgrade":

            @functools.wraps(func)
            def wrapper(
                path: str,
                revision: str,
                selector: str,
                url: str,
                *args,
                **kwargs,
            ) -> None:
                """
                Adapts Manager's db_upgrade method to accept cli arguments.

                :param path: Migration directory path
                :param revision: Parent revision
                :param selector: Database name in settings.py module
                :param url: SQLAlchemy Database URL
                :return: None
                """
                try:
                    config = Assistant.provide_alembic_config(
                        path, selector, url
                    )
                    func(config, revision, selector, *args, **kwargs)
                except NoSuchModuleError:
                    raise BasicManagerException(
                        "Command upgrade can only be performed on SQL "
                        f"database, {selector} is not supported SQL database"
                    )

        else:
            logger.warning(f"Decorated command {command} is not supported")

            @functools.wraps(func)
            def wrapper(*args, **kwargs) -> None:
                """
                Does nothing.

                :return: None
                """
                func(*args, **kwargs)

        return wrapper

    return decorator



from .adapter import IAdapter
from .assistant import IAssistant
from .exporter import IExporter
from .finding import IFinding
from .manager import IManager
from .observation import IObservation
from .observer import IObserver



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import AsyncExitStack, asynccontextmanager
from typing import Any, Callable, Optional, Union

from aiofile import async_open
from aiofile.utils import FileIOWrapperBase
from loguru import logger

from illuminate.meta.type import Result
from illuminate.observation import Observation


class FileObservation(Observation):
    """
    FileObservation class, reads file content asynchronously. Inherits
    Observation class and implements observe method.
    """

    def __hash__(self) -> int:
        """
        FileObservation object hash value.

        :return: int
        """
        return hash(self.url)

    def __init__(
        self,
        url: str,
        /,
        callback: Callable[[FileIOWrapperBase, tuple, dict], Result],
        xcom: Optional[Any] = None,
        *args,
        **kwargs,
    ):
        """
        FileObservation's __init__ method.

        :param url: File path
        :param callback: Async function/method that manipulates
        FileIOWrapperBase object and returns Result
        :param xcom: Cross communication object
        """
        super().__init__(url, xcom=xcom)
        self._callback = callback

    @asynccontextmanager
    async def observe(
        self, *args, **kwargs
    ) -> AsyncIterator[Union[None, Result]]:
        """
        Opens IO file stream asynchronously, pass stream object to a callback
        and returns None or Result as a context manager.

        :return: AsyncIterator with None or Result
        """
        _file = None
        _items = None
        async with AsyncExitStack() as stack:
            try:
                _file = await stack.enter_async_context(
                    async_open(self.url, "r")
                )
                logger.info(f"{self}.observe() -> {_file}")
                _items = self._callback(_file, *args, **kwargs)
            except Exception as exception:
                logger.warning(f"{self}.observe() -> {exception}")
            finally:
                yield _items
                if _file:
                    await _file.close()

    def __repr__(self):
        """
        FileObservation's __repr__ method.

        :return: String representation of an instance
        """
        return f'FileObservation("{self.url}",callback="{self._callback}")'



from __future__ import annotations

from typing import Any, Optional

from illuminate.exceptions import BasicObservationException
from illuminate.interface import IObservation


class Observation(IObservation):
    """
    Observation class, reads data from the source. Class must be inherited and
    method observe must be implemented in a child class.
    """

    def __hash__(self):
        """
        Observation object hash value.

        :return: None
        :raises BasicObservationException:
        """
        raise BasicObservationException(
            "Property hash must be implemented in child class"
        )

    def __init__(self, url: Any, xcom: Optional[Any] = None):
        """
        Observation's __init__ method.

        :param url: Data's URL
        :param xcom: Cross communication object
        """
        self.url = url
        self.xcom = xcom

    async def observe(self, *args, **kwargs):
        """
        Reads data from the source. Must be implemented in a child class.

        :return: None
        :raises BasicObservationException:
        """
        raise BasicObservationException(
            "Method observe must be implemented in child class"
        )



from __future__ import annotations

from typing import Type, TypeVar, Union

from loguru import logger
from sqlalchemy.ext.asyncio import AsyncSession

from illuminate.exceptions import BasicExporterException
from illuminate.exporter import Exporter

M = TypeVar("M")


class SQLExporter(Exporter):
    """
    SQLExporter class, writes data to SQL database asynchronously. Inherits
    Exporter class and implements export method.

    Each SQLExporter object is responsible for a single transaction with a
    single database. Attribute name is used to acquire database session object
    from Manager's sessions attribute.

    Supported dialects:
        - Mysql
        - Postgres
    """

    name: str
    """
    SQL database name selector used to get database session object from
    Manager.sessions attribute.
    """

    def __init__(self, models: Union[list[M], tuple[M]]):
        """
        SQLExporter's __init__ method.

        :param models: SQLAlchemy model objects collection
        """
        self.models = models

    async def export(
        self, session: Type[AsyncSession], *args, **kwargs
    ) -> None:
        """
        Writes data to SQL database asynchronously.

        :param session: AsyncSession object
        :return: None
        :raises BasicExporterException:
        """
        async with session() as session:  # type: ignore
            async with session.begin():  # type: ignore
                session.add_all(self.models)  # type: ignore
                try:
                    await session.commit()  # type: ignore
                except Exception as exception:
                    logger.warning(
                        f'{self}.export(session="{session}") -> {exception}'
                    )
                    raise BasicExporterException
        logger.success(f'{self}.export(session="{session}")')

    def __repr__(self):
        """
        SQLExporter's __repr__ method.

        :return: String representation of an instance
        """
        return f"SQLExporter(models={self.models})"



class IAdapter:
    """Interface for Adapter class."""

    async def adapt(self, finding, *args, **kwargs):
        """Generates Exporter and Observation objects."""
        raise NotImplementedError



_ADAPTER_EXAMPLE = """
from __future__ import annotations

from collections.abc import AsyncGenerator
from typing import Type, Union

from illuminate.adapter import Adapter
from illuminate.observation import FileObservation
from illuminate.observation import HTTPObservation
from illuminate.observation import SQLObservation
from illuminate.observation import SplashObservation
from illuminate.observer import Finding

from exporters.example import ExporterInfluxDBExample
from exporters.example import ExporterSQLExample
from findings.example import FindingExample
from models.example import ModelExample


class AdapterExample(Adapter):
    \"\"\"
    Adapter class is responsible for turning Finding objects into Exporter or
    Observation objects, and yielding them when adapt method is called.

    It is also designated to be a place where any additional enrichment of data
    should be performed, calling external services with some async library. If
    additional data can be used to construct URL, additional Observations can
    be yielded. For more information how to yield Observation object, check
    {name}/observers/example.py.

    Attribute subscribers is a collection of Finding classes that will be
    processed by Adapter. If Finding is subscribed to two or more Adapters,
    priority in adaptation will be give to Adapter with higher priority score.

    Manager's instance (object that is running whole process) is passed as
    an argument when initializing Adapters. This allows Adapter object to
    interact with Manager object attributes, like database sessions or
    queues for advanced ETL flows, by simply asking for
    self.manager.sessions["postgresql"]["main"] to acquire async database
    session defined in {name}/settings.py.

    Note: Method adapt can not yield Findings.
    \"\"\"

    priority: int = 10
    subscribers: tuple[Type[Finding]] = (FindingExample,)

    async def adapt(
        self, finding: FindingExample, *args, **kwargs
    ) -> AsyncGenerator[
        Union[
            ExporterInfluxDBExample,
            ExporterSQLExample,
            FileObservation,
            HTTPObservation,
            SQLObservation,
            SplashObservation,
        ],
        None,
    ]:
        yield ExporterSQLExample(
            models=[
                ModelExample(
                    load_time=finding.load_time,
                    title=finding.title,
                    url=finding.url
                )
            ]
        )

        yield ExporterInfluxDBExample(
            points={{
                "measurement": "{name}",
                "tags": {{"url": finding.url, "title": finding.title}},
                "fields": {{"load_time": finding.load_time}},
            }}
        )

"""

_ALEMBIC_ENV_PY = """
from pydoc import locate

from alembic import context
from sqlalchemy import engine_from_config
from sqlalchemy import pool

from models import Base
from settings import MODELS

for model in MODELS:
    locate(model)

config = context.config


def run_migrations_offline():
    url = config.get_main_option("sqlalchemy.url")
    context.configure(
        url=url,
        target_metadata=Base.metadata,
        literal_binds=True,
        dialect_opts={{"paramstyle": "named"}},
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    connectable = engine_from_config(
        config.get_section(config.config_ini_section),
        prefix="sqlalchemy.",
        poolclass=pool.NullPool,
    )

    with connectable.connect() as connection:
        context.configure(
            connection=connection, target_metadata=Base.metadata
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()

"""

_ALEMBIC_SCRIPT_PY_MAKO = """
\"\"\"${{message}}

Revision ID: ${{up_revision}}
Revises: ${{down_revision | comma,n}}
Create Date: ${{create_date}}

\"\"\"
from alembic import op
import sqlalchemy as sa
${{imports if imports else ""}}

# revision identifiers, used by Alembic.
revision = ${{repr(up_revision)}}
down_revision = ${{repr(down_revision)}}
branch_labels = ${{repr(branch_labels)}}
depends_on = ${{repr(depends_on)}}


def upgrade():
    ${{upgrades if upgrades else "pass"}}


def downgrade():
    ${{downgrades if downgrades else "pass"}}

"""

_DOCKER_COMPOSE = """
version: '3.8'
services:
  grafana:
   image: grafana/grafana
   container_name: grafana
   restart: always
   depends_on:
     - influxdb
   environment:
     - GF_SECURITY_ADMIN_USER=illuminate
     - GF_SECURITY_ADMIN_PASSWORD=$ILLUMINATE_GRAFANA_PASSWORD
     - GF_INSTALL_PLUGINS=
   links:
     - influxdb
   ports:
     - '3000:3000'
   volumes:
     - grafana:/var/lib/grafana
  influxdb:
    image: influxdb:1.8
    container_name: influxdb
    restart: always
    environment:
      - INFLUXDB_ADMIN_USER=illuminate
      - INFLUXDB_ADMIN_PASSWORD=$ILLUMINATE_MEASUREMENTS_DB_PASSWORD
      - INFLUXDB_DB={name}
      - INFLUXDB_HTTP_AUTH_ENABLED=true
    ports:
      - '8086:8086'
    volumes:
      - influxdb:/var/lib/influxdb
  pg:
    container_name: pg
    image: postgres:latest
    restart: always
    environment:
      - POSTGRES_USER=illuminate
      - POSTGRES_PASSWORD=$ILLUMINATE_MAIN_DB_PASSWORD
      - POSTGRES_DB={name}
    ports:
      - '5432:5432'
    volumes:
      - postgres:/var/lib/postgresql/data
  pgadmin:
    container_name: pgadmin
    image: dpage/pgadmin4
    restart: always
    environment:
      - PGADMIN_DEFAULT_EMAIL=root@example.com
      - PGADMIN_DEFAULT_PASSWORD=$ILLUMINATE_PGADMIN_PASSWORD
    ports:
      - "8080:80"
  splash:
    container_name: splash
    image: scrapinghub/splash
    restart: always
    ports:
      - "8050:8050"
volumes:
  grafana:
    driver: local
  influxdb:
    driver: local
  postgres:
    driver: local

"""

_EMPTY = """
"""

_EXPORTER_EXAMPLE = """
from __future__ import annotations

from typing import Iterable, Union

from illuminate.exporter import InfluxDBExporter
from illuminate.exporter import SQLExporter
from pandas import DataFrame

from models.example import ModelExample


class ExporterInfluxDBExample(InfluxDBExporter):
    \"\"\"
    InfluxDBExporter class will write points to database using session. Points
    are passed at initialization, while database session is found by name
    attribute in the pool of existing sessions. Name must co-respond to DB
    section in {name}/settings.py. For more information how to initialize
    InfluxDBExporter class, check {name}/adapters/example.py
    \"\"\"

    name: str = "measurements"

    def __init__(
        self,
        points: Union[
            Union[DataFrame, dict, str, bytes],
            Iterable[Union[DataFrame, dict, str, bytes]],
        ],
    ):
        super().__init__(points)


class ExporterSQLExample(SQLExporter):
    \"\"\"
    SQLExporter class will commit models to database using session. Models are
    passed at initialization, while database session is found by name attribute
    in the pool of existing sessions. Name must co-respond to DB section in
    {name}/settings.py. For more information how to initialize SQLExporter
    class, check {name}/adapters/example.py
    \"\"\"

    name: str = "main"

    def __init__(self, models: Union[list[ModelExample], tuple[ModelExample]]):
        super().__init__(models)

"""

_FIXTURE_EXAMPLE = """
[
    {{
    "name": "{name}",
    "data": [
        {{
            "load_time": "1.0",
            "url": "https://webscraper.io/",
            "title": "Web Scraper - The #1 web scraping extension"
        }},
        {{
            "load_time": "1.0",
            "url": "https://webscraper.io/tutorials",
            "title": "Web Scraper Tutorials"
        }}
    ]
    }}
]
"""

_MODELS = """
from sqlalchemy.orm import declarative_base

Base = declarative_base()

"""

_MODEL_EXAMPLE = """
from sqlalchemy import Column, Float, Integer, String

from models import Base


class ModelExample(Base):
    \"\"\"
    SQLAlchemy model used by SQLExporter object. For more information about
    SQLExporter class check {name}/exporters/example.py.
    \"\"\"

    __tablename__ = "{name}"
    id: int = Column(Integer, primary_key=True)
    load_time: float = Column(Float)
    title: str = Column(String)
    url: str = Column(String)

    def __repr__(self):
        \"\"\"ModelExample's __repr__ method.\"\"\"
        return (
            f'ModelExample(load_time={{self.load_time}},'
            f'title="{{self.title}}",url="{{self.url}}")'
        )

"""

_ILLUMINATE_SETTINGS = """
\"\"\"
This file represents project {name}'s settings. It will be imported by a
framework and used to configure run. The sections are as following:

* CONCURRENCY
Number of workers per queue type. I/O heavy queues can have more workers
assigned to them to exploit longer wait times.

* DB
Database related data used by SQLAlchemy to acquire sessions. Sessions are
obtained at the start of the ETL process and can be accessed by instantiating
Manager class and access sessions attribute.

* MODELS
List of SQLAlchemy models affected by illuminate cli when invoking
'db revision', 'db upgrade' and 'db populate' commands.

* NAME
Project name.

* OBSERVATION_CONFIGURATION
General and Observation type specific configuration. Type specific
configuration is used if Observation is not specifying its own.
\"\"\"

import os

from illuminate import __version__

CONCURRENCY = {{
    "observers": 8,
    "adapters": 2,
    "exporters": 8,
}}

DB = {{
    "main": {{
        "host": "localhost",
        "name": "{name}",
        "pass": os.environ.get("ILLUMINATE_MAIN_DB_PASSWORD"),
        "port": "5432",
        "user": "illuminate",
        "type": "postgresql",
    }},
    "measurements": {{
        "host": "localhost",
        "name": "{name}",
        "pass": os.environ.get("ILLUMINATE_MEASUREMENTS_DB_PASSWORD"),
        "port": "8086",
        "user": "illuminate",
        "type": "influxdb",
    }},
}}

MODELS = [
    "models.example.ModelExample",
]

NAME = "{name}"

OBSERVATION_CONFIGURATION = {{
    "delay": 0.1,
    "http": {{
        "auth_username": None,
        "auth_password": None,
        "connect_timeout": 10.0,
        "body": None,
        "headers": None,
        "method": "GET",
        "request_timeout": 10.0,
        "user_agent": f"Illuminate-bot/{{__version__}}",
        "validate_cert": False,
    }},
    "splash": {{
        "body": "",
        "host": "localhost",
        "method": "GET",
        "port": 8050,
        "protocol": "http",
        "render": "html",
        "timeout": 30,
    }}
}}

"""

_FINDING_EXAMPLE = """
from dataclasses import dataclass
from dataclasses import field

from illuminate.observer import Finding


@dataclass(frozen=True, order=True)
class FindingExample(Finding):
    \"\"\"
    Finding is a data class, meant to hold raw data extracted by Observation's
    callback. Finding will be passed to Adapter object's adapt method if
    it is subscribed to Adapter.

    Check {name}/adapters/example.py to learn more about subscription.
    \"\"\"

    load_time: float = field()
    title: str = field()
    url: str = field()

"""

_OBSERVER_EXAMPLE = """
from __future__ import annotations

from collections.abc import AsyncGenerator, Generator
from typing import Optional, Union
from urllib.parse import urldefrag
from urllib.parse import urljoin

from bs4 import BeautifulSoup
from illuminate.manager import Manager
from illuminate.observation import HTTPObservation
from illuminate.observer import Observer
from tornado.httpclient import HTTPResponse

from exporters.example import ExporterInfluxDBExample
from exporters.example import ExporterSQLExample
from findings.example import FindingExample


def _create_soup(response: HTTPResponse) -> BeautifulSoup:
    html = response.body.decode(errors="ignore")
    soup = BeautifulSoup(html, "html.parser")
    return soup


def _extract_hrefs(
    soup: BeautifulSoup, url: str
) -> Generator[str, None]:
    links = soup.find_all("a")
    for link in links:
        _url = link.get("href")
        if _url:
            yield urljoin(url, urldefrag(_url)[0])


class ObserverExample(Observer):
    \"\"\"
    Observer is ETL configuration class. It represents entry point, and it
    defines the flow with observe, and any additional method.

    Once initialized by the framework, it will fill an observation queue with
    objects from Observer's initial_observation collection, starting the whole
    process.

    Note: Multiple Observers can exist in the same project.
    \"\"\"

    ALLOWED: Union[list[str], tuple[str]] = ("https://webscraper.io/",)
    \"\"\"
    Collection of strings evaluated against URL to determent if URL is allowed
    to be observed. If empty, no Observation wide restrictions are forced.
    \"\"\"
    NAME: str = "example"
    \"\"\"Observer's name.\"\"\"

    def __init__(self, manager: Optional[Manager] = None):
        \"\"\"
        Collection initial_observations is de facto entry point. It must
        contain Observation objects initialized with URL, allowed list of
        strings and callback method. If Observation's URL starts with element
        in allowed collection, it will be performed. Otherwise, it is
        rejected.

        Manager's instance (object that is running whole process) is passed as
        an argument when initializing Observers. This allows Observer object to
        interact with Manager object attributes, like database sessions or
        queues for advanced ETL flows, by simply asking for
        self.manager.sessions["postgresql"]["main"] to acquire async database
        session defined in {name}/settings.py.
        \"\"\"

        super().__init__(manager)
        self.initial_observations = [
            HTTPObservation(
                "https://webscraper.io/",
                allowed=self.ALLOWED,
                callback=self.observe,
            ),
        ]

    async def observe(
        self, response: HTTPResponse, *args, **kwargs
    ) -> AsyncGenerator[
        Union[
            ExporterInfluxDBExample,
            ExporterSQLExample,
            FindingExample,
            HTTPObservation,
        ],
        None,
    ]:
        \"\"\"
        ETL flow is regulated by a yielded object type of Observation's
        callback method. Each object type corresponds to ETL stage:

        * Observation -> Extract
        * Finding -> Transform
        * Exporter -> Load

        In the example below, initial HTTPObservation returned Tornado's
        HTTP response object that was used to extract all hrefs from HTML.
        These hrefs will be used as URLs for new HTTPObservations, using
        the same observe method as a callback and same allowed collection.
        Finally, Finding object is yielded, representing a desired data.

        This flow, with everything set, represents a simple web scraper that
        will visit every page found on a domain, and take page's title and URL
        as desired data.

        Tip: If there is no need for some further data enrichment or
        manipulation, yield Exporter object instead of Finding object.
        For reference check {name}/adapters/example.py.

        Note: Illuminate takes care of tracking what resources were already
        requested, avoiding duplication. Resource check pool is shared between
        Observers, avoiding overlapping.
        \"\"\"

        soup = _create_soup(response)
        hrefs = _extract_hrefs(soup, response.effective_url)
        for href in hrefs:
            yield HTTPObservation(
                href,
                allowed=self.ALLOWED,
                callback=self.observe,
            )
        yield FindingExample(
            response.request_time,
            soup.title.text,
            response.effective_url
        )

"""

FILES = {
    "docker-compose.yaml": _DOCKER_COMPOSE,
    "settings.py": _ILLUMINATE_SETTINGS,
    "adapters/__init__.py": _EMPTY,
    "adapters/example.py": _ADAPTER_EXAMPLE,
    "exporters/__init__.py": _EMPTY,
    "exporters/example.py": _EXPORTER_EXAMPLE,
    "fixtures/example.json": _FIXTURE_EXAMPLE,
    "models/__init__.py": _MODELS,
    "models/example.py": _MODEL_EXAMPLE,
    "migrations/env.py": _ALEMBIC_ENV_PY,
    "migrations/script.py.mako": _ALEMBIC_SCRIPT_PY_MAKO,
    "migrations/versions/.gitkeep": _EMPTY,
    "findings/__init__.py": _EMPTY,
    "observers/__init__.py": _EMPTY,
    "findings/example.py": _FINDING_EXAMPLE,
    "observers/example.py": _OBSERVER_EXAMPLE,
}



from __future__ import annotations

import os
import sys
import warnings
from unittest.mock import patch

import click
from loguru import logger

from illuminate import __version__
from illuminate.common import LOGGING_LEVELS
from illuminate.manager import Assistant
from illuminate.manager import Manager


@click.group()
@click.version_option(__version__)
@click.option(
    "--verbosity",
    default=LOGGING_LEVELS[2],
    help="Configure logging levels.",
    required=False,
    type=click.Choice(LOGGING_LEVELS),
    show_default=True,
)
def cli(verbosity: str) -> None:
    """Framework entrypoint."""
    logger.remove()
    logger.add(sys.stdout, level=verbosity)
    sys.path.insert(0, os.getcwd())
    if not sys.warnoptions:
        warnings.simplefilter("ignore")


@cli.group("manage")
def manage() -> None:
    """Framework manage group of commands."""
    pass


@cli.group("observe")
def observe() -> None:
    """Framework observe group of commands."""
    pass


@manage.group("db")
def db() -> None:
    """Prepares relational db for ETL operations."""
    pass


@manage.group("project")
def project() -> None:
    """Performs project operations."""
    pass


@db.command("populate")
@click.option(
    "--fixtures",
    help="Fixture files paths.",
    multiple=True,
    required=True,
    type=click.Path(exists=True),
)
@click.option(
    "--selector",
    default="main",
    help="Database connection selector.",
    required=False,
    show_default=True,
    type=str,
)
@click.option(
    "--url",
    default=None,
    help="Optional URL for databases not included in settings module.",
    required=False,
    type=str,
)
def db_populate(fixtures: tuple[str], selector: str, url: str) -> None:
    """Populates database with fixtures."""
    Manager.db_populate(fixtures, selector, url)


@db.command("revision")
@click.option(
    "--revision",
    default="head",
    help="Alembic revision selector.",
    required=False,
    show_default=True,
    type=str,
)
@click.option(
    "--selector",
    default="main",
    help="Database connection selector.",
    required=False,
    show_default=True,
    type=str,
)
@click.option(
    "--url",
    default=None,
    help="Optional URL for databases not included in settings module.",
    required=False,
    type=str,
)
@click.argument(
    "path", default=os.getcwd(), required=False, type=click.Path(exists=True)
)
def db_revision(
    path: str,
    revision: str,
    selector: str,
    url: str,
) -> None:
    """Creates Alembic's revision file in migration directory."""
    with patch("alembic.util.messaging.write_outstream"):
        Manager.db_revision(path, revision, selector, url)


@db.command("upgrade")
@click.option(
    "--revision",
    default="head",
    help="Alembic revision selector.",
    required=False,
    show_default=True,
    type=str,
)
@click.option(
    "--selector",
    default="main",
    help="Database connection selector.",
    required=False,
    show_default=True,
    type=str,
)
@click.option(
    "--url",
    default=None,
    help="Optional URL for databases not included in settings module.",
    required=False,
    type=str,
)
@click.argument(
    "path", default=os.getcwd(), required=False, type=click.Path(exists=True)
)
def db_upgrade(
    path: str,
    revision: str,
    selector: str,
    url: str,
) -> None:
    """Applies migration file to a database."""
    Manager.db_upgrade(path, revision, selector, url)


@project.command("setup")
@click.argument("name", required=True, type=str)
@click.argument(
    "path", default=os.getcwd(), required=False, type=click.Path(exists=True)
)
def setup(name: str, path: str) -> None:
    """Creates a project directory with all needed files."""
    Manager.project_setup(name, path)


@observe.command("catalogue")
def catalogue() -> None:
    """Lists observers found in project files."""
    Manager.observe_catalogue(**Assistant.provide_context(sessions=False))


@observe.command("start")
@click.option(
    "--observer",
    help="Observer selector. Leave empty to include all observers.",
    multiple=True,
    required=False,
    type=str,
)
def start(observer: tuple[str]) -> None:
    """Starts producer/consumer ETL process."""
    context = Assistant.provide_context(_filter=observer)
    manager = Manager(**context)  # type: ignore
    manager.observe_start()


if __name__ == "__main__":
    cli(obj={})



from __future__ import annotations

from typing import AsyncGenerator, Coroutine, Union

from illuminate.exporter import Exporter
from illuminate.observation import Observation
from illuminate.observer import Finding

Result = Union[
    AsyncGenerator[Union[Exporter, Finding, Observation], None],
    Coroutine,
]



from __future__ import annotations

from typing import Any, Callable, Optional, Type, Union

from loguru import logger
from sqlalchemy.engine.result import Result as AlchemyResult
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.sql.elements import TextClause
from sqlalchemy.sql.selectable import Select

from illuminate.meta.type import Result
from illuminate.observation import Observation


class SQLObservation(Observation):
    """
    SQLObservation class, reads data from database asynchronously. Inherits
    Observation class and implements observe method.
    """

    def __hash__(self) -> int:
        """
        SQLObservation object hash value.

        :return: int
        """
        query = str(self.query)
        return hash(f"{self.url}|:{query}")

    def __init__(
        self,
        query: Union[Select, TextClause],
        url: str,
        /,
        callback: Callable[[AlchemyResult, tuple, dict], Result],
        xcom: Optional[Any] = None,
        *args,
        **kwargs,
    ):
        """
        SQLObservation's __init__ method.

        :param query: SQLAlchemy query object.
        :param url: Database name in project settings.
        :param callback: Async function/method that manipulates AlchemyResult
        object and returns Result.
        :param xcom: Cross communication object
        """
        super().__init__(url, xcom=xcom)
        self._callback = callback
        self.query = query

    async def observe(
        self, session: Type[AsyncSession], *args, **kwargs
    ) -> Union[None, Result]:
        """
        Reads data from database, passes response object to a callback
        and returns None or Result.

        :return: None or Result
        """
        try:
            async with session() as session:  # type: ignore
                async with session.begin():  # type: ignore
                    query = self.query
                    response = await session.execute(query)  # type: ignore
            logger.info(f'{self}.observe(session="{session}")')
            return self._callback(response, *args, **kwargs)
        except Exception as exception:
            logger.warning(f"{self}.observe() -> {exception}")
            return None

    def __repr__(self):
        """
        SQLObservation's __repr__ method.

        :return: String representation of an instance
        """
        return (
            f'SQLObservation("{self.query}","{self.url}",'
            f'callback="{self._callback}")'
        )



class IManager:
    """Interface for Manager class."""

    @staticmethod
    def db_populate(fixtures, models, operations, selector):
        """Populates database with fixtures."""
        raise NotImplementedError

    @staticmethod
    def db_revision(config, revision):
        """Creates Alembic's revision file in migration directory."""
        raise NotImplementedError

    @staticmethod
    def db_upgrade(config, revision, selector):
        """Applies migration file to a database."""
        raise NotImplementedError

    @staticmethod
    def project_setup(name, path):
        """Creates a project directory with all needed files."""
        raise NotImplementedError

    def observe_start(self):
        """Starts producer/consumer ETL process."""
        raise NotImplementedError



class IAssistant:
    """Interface for Assistant class."""

    @staticmethod
    def provide_alembic_config(path, selector, url):
        """Creates Alembic's configuration object."""
        raise NotImplementedError

    @staticmethod
    def provide_alembic_operations(selector, url):
        """Creates Alembic's operations object."""
        raise NotImplementedError

    @staticmethod
    def provide_context(sessions, _filter):
        """Creates Manager's constructor kwargs."""
        raise NotImplementedError

    @staticmethod
    def provide_models():
        """Gathers project models."""
        raise NotImplementedError

    @staticmethod
    def _provide_db_url(selector):
        """Creates database URL."""
        raise NotImplementedError

    @staticmethod
    def _provide_sessions():
        """Creates a dictionary of database sessions."""
        raise NotImplementedError

    @staticmethod
    def _provide_settings():
        """Imports project's settings.py module and returns it."""
        raise NotImplementedError



from illuminate.exceptions import BasicExporterException
from illuminate.interface import IExporter


class Exporter(IExporter):
    """
    Exporter class, writes data to destination. Class must be inherited and
    method export must be implemented in a child class.
    """

    async def export(self, *args, **kwargs):
        """
        Writes data to destination. Must be implemented in a child class.

        :return: None
        :raises BasicExporterException:
        """
        raise BasicExporterException(
            "Method export must be implemented in child class"
        )
